from .indexes import INDEXES
from .lists import STOCK_LISTS
from py_portfolio_index.models import StockInfo
from pathlib import Path
from json import loads
from pydantic import __version__
from packaging import version

# legacy handling
if version.parse(__version__) < version.parse("2.0.0"):
    setattr(StockInfo, "model_validate", StockInfo.parse_obj)


STOCK_INFO: dict[str, StockInfo] = {}
with open(Path(__file__).parent / "stock_info.json", "r", encoding="utf-8") as f:
    content = f.read()
    if content:
        all = loads(content)
        for row in all:
            STOCK_INFO[row["ticker"]] = StockInfo.model_validate(row)

with open(Path(__file__).parent / "cached_ticker_list.csv", "r") as f:
    VALID_STOCKS = set([v for v in f.read().split("\n") if v])

__all__ = ["INDEXES", "STOCK_LISTS", "STOCK_INFO", "VALID_STOCKS"]



from dataclasses import dataclass
from py_portfolio_index.enums import Currency, Provider
from typing import List


@dataclass
class Config:
    default_currency = Currency.USD


def get_providers() -> List[Provider]:
    providers = []
    try:
        from alpaca.trading.client import TradingClient  # noqa: F401

        providers.append(Provider.ALPACA)
        providers.append(Provider.ALPACA_PAPER)
    except ImportError:
        pass
    try:
        import robin_stocks.robinhood as r  # noqa: F401

        providers.append(Provider.ROBINHOOD)
    except ImportError:
        pass
    try:
        from webull import webull  # noqa: F401

        providers.append(Provider.WEBULL)
        providers.append(Provider.WEBULL_PAPER)
    except ImportError:
        pass
    return providers



from math import floor, ceil
from typing import Dict, Union, Optional, Set, List, Callable, Any, DefaultDict
from decimal import Decimal
from datetime import date

from py_portfolio_index.common import (
    print_money,
    print_per,
    round_up_to_place,
    get_basic_stock_info,
)
from py_portfolio_index.constants import Logger
from py_portfolio_index.enums import RoundingStrategy, Provider
from py_portfolio_index.exceptions import PriceFetchError, OrderError
from py_portfolio_index.models import (
    Money,
    OrderPlan,
    OrderElement,
    StockInfo,
    ProfitModel,
)
from functools import lru_cache
from py_portfolio_index.models import RealPortfolio
from dataclasses import dataclass, field
from enum import Enum
from datetime import datetime


@dataclass
class CachedValue:
    value: Any
    fetcher: Callable
    set: datetime = field(default_factory=datetime.now)


class CacheKey(Enum):
    POSITIONS = 0
    DIVIDENDS = 1
    UNSETTLED = 2
    ACCOUNT = 3
    OPEN_ORDERS = 4
    MISC = 5


class BaseProvider(object):
    PROVIDER: Provider = Provider.DUMMY
    MIN_ORDER_VALUE = Money(value=1)
    MAX_ORDER_DECIMALS = 2
    SUPPORTS_BATCH_HISTORY = 0
    CACHE: dict[str, CachedValue] = {}

    def clear_cache(self, skip_clearing: List[str]):
        for value in self.CACHE.values():
            if value in skip_clearing:
                continue
            value.value = None

    def _get_cached_value(
        self,
        key: CacheKey,
        value: Optional[Any] = None,
        max_age_seconds: int = 60 * 60,
        callable: Optional[Callable] = None,
    ) -> Any:
        if value:
            skey = f"{key}_{value}"
        else:
            skey = f"{key}"
        if skey in self.CACHE:
            cached = self.CACHE[skey]
        elif callable:
            cached = CachedValue(value=None, fetcher=callable)
        if cached.value:
            age = datetime.now() - cached.set
            if age.seconds < max_age_seconds:
                return cached.value
        cached.value = cached.fetcher()
        return cached.value

    def __init__(self) -> None:
        self.stock_info_cache: Dict[str, StockInfo] = {}

    @property
    def valid_assets(self) -> Set[str]:
        return set()

    def _get_instrument_price(self, ticker: str, at_day: Optional[date] = None):
        raise NotImplementedError

    def get_holdings(self) -> RealPortfolio:
        raise NotImplementedError

    def get_per_ticker_profit_or_loss(self) -> Dict[str, ProfitModel]:
        raise NotImplementedError

    def get_profit_or_loss(self) -> ProfitModel:
        raw = self.get_per_ticker_profit_or_loss().values()
        appreciation = sum([x.appreciation for x in raw], Money(value=0.0))
        dividends = sum([x.dividends for x in raw], Money(value=0.0))
        return ProfitModel(appreciation=appreciation, dividends=dividends)

    def get_instrument_prices(
        self, tickers: List[str], at_day: Optional[date] = None
    ) -> Dict[str, Optional[Decimal]]:
        output = {}
        for ticker in tickers:
            output[ticker] = self.get_instrument_price(ticker, at_day=at_day)
        return output

    @lru_cache(maxsize=None)
    def get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        try:
            return self._get_instrument_price(ticker, at_day)
        except NotImplementedError as e:
            raise e
        except Exception as e:
            raise PriceFetchError(e)

    def buy_instrument(
        self, ticker: str, qty: Decimal, value: Optional[Money] = None
    ) -> bool:
        raise NotImplementedError

    def get_unsettled_instruments(self) -> Set[str]:
        raise NotImplementedError

    def purchase_ticker_value_dict(
        self,
        to_buy: Dict[str, Money],
        purchasing_power: Union[Money, Decimal, int, float],
        plan_only: bool = False,
        fractional_shares: bool = True,
        skip_errored_stocks=False,
        rounding_strategy=RoundingStrategy.CLOSEST,
        ignore_unsettled: bool = True,
    ):
        purchased = Money(value=0)
        purchasing_power_resolved = Money(value=purchasing_power)
        target_value: Money = Money(value=sum([v for k, v in to_buy.items()]))
        diff = Money(value=0)
        if ignore_unsettled:
            unsettled = self.get_unsettled_instruments()
        else:
            unsettled = set()
        break_flag = False
        for key, value in to_buy.items():
            if key in unsettled:
                Logger.info(f"Skipping {key} with unsettled orders.")
                continue
            try:
                raw_price = self.get_instrument_price(key)
                if not raw_price:
                    raise ValueError(f"No price found for this instrument: {key}")
                price: Money = Money(value=raw_price)
                Logger.info(f"got price of {price} for {key}")
            except Exception as e:
                if not skip_errored_stocks:
                    raise e
                else:
                    continue
            if not price:
                price = Money(value=0)
            if price == Money(value=0):
                to_buy_currency = Money(value=0)
            else:
                to_buy_currency = value / price

            if fractional_shares:
                to_buy_units = round(to_buy_currency, 4)
            else:
                if rounding_strategy == RoundingStrategy.CLOSEST:
                    to_buy_units = Money(value=int(round(to_buy_currency, 0)))
                elif rounding_strategy == RoundingStrategy.FLOOR:
                    to_buy_units = Money(value=floor(to_buy_currency))
                elif rounding_strategy == RoundingStrategy.CEILING:
                    to_buy_units = Money(value=ceil(to_buy_currency))
                else:
                    raise ValueError(
                        "Invalid rounding strategy provided with non-fractional shares."
                    )
            if not to_buy_units:
                Logger.info(f"skipping {key} because no units to buy")
                continue
            purchasing = to_buy_units * price

            Logger.info(f"Need to buy {to_buy_units} units of {key}.")
            if (purchasing_power_resolved - purchasing) < Money(value=0):
                Logger.info("Out of money, buying what is possible and exiting")
                break_flag = True
                purchasing = purchasing_power_resolved
                to_buy_units = Decimal(round(purchasing / price, 4).value)
            if to_buy_units > Decimal(0):
                Logger.info(f"going to buy {to_buy_units} of {key}")
                try:
                    if not plan_only:
                        successfully_purchased = self.buy_instrument(key, to_buy_units)
                    if successfully_purchased:
                        purchasing_power_resolved = (
                            purchasing_power_resolved - purchasing
                        )
                        purchased += purchasing
                        diff += abs(value - purchasing)
                        Logger.info(
                            f"bought {to_buy_units} of {key}, {purchasing_power_resolved} left"
                        )
                except Exception as e:
                    print(e)
                    if not skip_errored_stocks:
                        raise e
            if break_flag:
                Logger.info(
                    f"No purchasing power left, purchased {print_money(purchased)} of {print_money(target_value)}."
                )
                break
        Logger.info(
            f"$ diff from ideal for purchased stocks was {print_money(diff)}. {print_per(diff / target_value)} of total purchase goal."
        )

    def handle_order_element(self, element: OrderElement, dry_run: bool = False):
        raw_price = self.get_instrument_price(element.ticker)

        if not raw_price:
            raise OrderError(f"No price found for this instrument: {element.ticker}")
        price: Money = Money(value=raw_price)
        if element.qty:
            units = Decimal(element.qty)
            value = Money(value=units * price.decimal)

        elif element.value:
            raw_price = self.get_instrument_price(element.ticker)
            Logger.info(f"got price of {price} for {element.ticker}")
            units = round_up_to_place(
                (element.value / price).decimal, self.MAX_ORDER_DECIMALS
            )
            value = element.value
        else:
            raise OrderError("Order element must have qty or value")
        if not dry_run:
            self.buy_instrument(element.ticker, units, value)
            Logger.info(f"Bought {units} of {element.ticker}")
        else:
            Logger.info(f"Would have bought {units} of {element.ticker}")

    def _get_stock_info(self, ticker: str) -> dict:
        raise NotImplementedError

    def get_stock_info(self, ticker: str) -> StockInfo:
        cached = self.stock_info_cache.get(ticker, None)
        if not cached:
            dynamic = self._get_stock_info(ticker)
            basic = get_basic_stock_info(ticker, fail_on_missing=False)
            if basic:
                raw_data = basic.dict()
            else:
                raw_data = {}
            final = StockInfo(**{**raw_data, **dynamic})
            self.stock_info_cache[ticker] = final
            return final
        return cached

    def purchase_order_plan(
        self,
        plan: OrderPlan,
        skip_errored_stocks=False,
        ignore_unsettled: bool = True,
        plan_only: bool = False,
    ):
        if ignore_unsettled:
            unsettled = self.get_unsettled_instruments()
        else:
            unsettled = set()
        for item in plan.to_buy:
            if item.ticker in unsettled:
                Logger.info(f"Skipping {item.ticker} with unsettled orders.")
                continue
            try:
                self.handle_order_element(item, dry_run=plan_only)
            except Exception as e:
                print(e)
                if not skip_errored_stocks:
                    raise e

    def refresh(self):
        pass

    def _get_dividends(self) -> DefaultDict[str, Money]:
        raise NotImplementedError



# Class implementing a cache
# for historic stock prices
# requires a callable interace to get values at date for a list of tickers
# accepts list of stocks + dates to get values for
# returns these from cache if possible, or for those not found
# calls provider to return prices
from collections import defaultdict
from typing import List, Dict
from datetime import date as datetype
from datetime import datetime
from decimal import Decimal


class PriceCache(object):
    def __init__(self, fetcher):
        self.fetcher = fetcher
        self.store = defaultdict(dict)
        self.instant_refresh_times: dict[str, datetime] = {}
        self.default_timeout: int = 60 * 60  # 1 hour

    def get_prices(
        self, tickers: List[str], date: datetype | None = None
    ) -> Dict[str, Decimal]:
        # if no date is provided, assume they want the instantaneous price
        if not date:
            label = "INSTANT"
        else:
            label = date.isoformat()
        cached = self.store[label]
        found = {k: v for k, v in cached.items() if k in tickers}
        if label == "INSTANT":
            for k, v in self.instant_refresh_times.items():
                if k in tickers:
                    if (datetime.now() - v).seconds > self.default_timeout:
                        found.pop(k, None)
        missing = [x for x in tickers if x not in found]
        if missing:
            prices = self.fetcher(missing, date)
            for ticker, price in prices.items():
                cached[ticker] = price
                found[ticker] = price
                if label == "INSTANT":
                    self.instant_refresh_times[ticker] = datetime.now()
        return found



import re
from decimal import Decimal
from time import sleep
from datetime import date, datetime
from typing import Optional, List, Dict, DefaultDict
from py_portfolio_index.constants import Logger
from py_portfolio_index.models import (
    RealPortfolio,
    RealPortfolioElement,
    Money,
    ProfitModel,
)
from py_portfolio_index.common import divide_into_batches
from py_portfolio_index.portfolio_providers.common import PriceCache
from py_portfolio_index.portfolio_providers.base_portfolio import (
    BaseProvider,
    CacheKey,
)
from py_portfolio_index.exceptions import PriceFetchError, ConfigurationError
from py_portfolio_index.portfolio_providers.helpers.robinhood import (
    validate_login,
    ROBINHOOD_PASSWORD_ENV,
    ROBINHOOD_USERNAME_ENV,
)
from py_portfolio_index.enums import Provider
from functools import lru_cache
from os import environ

FRACTIONAL_SLEEP = 60
BATCH_SIZE = 50


def nearest_value(all_historicals, pivot) -> Optional[dict]:
    filtered = [z for z in all_historicals if z]
    if not filtered:
        return None
    return min(
        filtered,
        key=lambda x: abs(
            datetime.strptime(x["begins_at"], "%Y-%m-%dT%H:%M:%SZ").date() - pivot
        ),
    )


def nearest_multi_value(
    symbol: str, all_historicals, pivot: Optional[date] = None
) -> Optional[Decimal]:
    filtered = [z for z in all_historicals if z and z["symbol"] == symbol]
    if not filtered:
        return None
    if pivot is not None:
        lpivot = pivot or date.today()
        closest = min(
            filtered,
            key=lambda x: abs(
                datetime.strptime(x["begins_at"], "%Y-%m-%dT%H:%M:%SZ").date() - lpivot
            ),
        )
    else:
        closest = filtered[0]
    if closest:
        value = closest.get("last_trade_price", closest.get("high_price", None))
        return Decimal(value)
    return None


class InstrumentDict(dict):
    def __init__(self, refresher, *args):
        super().__init__(*args)
        self.refresher = refresher

    def __missing__(self, key):
        mapping = self.refresher()
        self.update(mapping)
        if key in self:
            return self[key]
        raise ValueError(f"Could not find instrument {key} after refresh")


class RobinhoodProvider(BaseProvider):
    """Provider for interacting with stocks held in
    Robinhood.
    """

    PROVIDER = Provider.ROBINHOOD
    SUPPORTS_BATCH_HISTORY = 70

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        skip_cache: bool = False,
        external_auth: bool = False,
    ):
        import robin_stocks.robinhood as r

        if not username:
            username = environ.get(ROBINHOOD_USERNAME_ENV, None)
        if not password:
            password = environ.get(ROBINHOOD_PASSWORD_ENV, None)
        if not (username and password):
            raise ConfigurationError(
                "Must provide username and password arguments or set environment variables ROBINHOOD_USERNAME and ROBINHOOD_PASSWORD "
            )
        self._provider = r
        BaseProvider.__init__(self)
        if not external_auth:
            self._provider.login(username=username, password=password)
        else:
            validate_login()
        self._local_instrument_cache: List[Dict] = []
        if not skip_cache:
            self._load_local_instrument_cache()
        self._local_latest_price_cache: Dict[str, Decimal] = {}
        self._price_cache: PriceCache = PriceCache(fetcher=self._get_instrument_prices)

    @property
    def valid_assets(self) -> set[str]:
        if not self._local_instrument_cache:
            self._load_local_instrument_cache()
        return self._get_cached_value(
            CacheKey.MISC,
            value="valid_tickers",
            callable=lambda: {row["symbol"] for row in self._local_instrument_cache},
        )

    def _load_local_instrument_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir("py_portfolio_index", ensure_exists=True))
        file = path / "robinhood_instruments.json"
        if not file.exists():
            self._local_instrument_cache = {}
            return
        with open(file, "r") as f:
            self._local_instrument_cache = json.load(f)

    def _save_local_instrument_cache(self):
        from platformdirs import user_cache_dir
        from pathlib import Path
        import json

        path = Path(user_cache_dir("py_portfolio_index", ensure_exists=True))
        file = path / "robinhood_instruments.json"
        with open(file, "w") as f:
            json.dump(self._local_instrument_cache, f)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        if at_day:
            historicals = self._provider.get_stock_historicals(
                [ticker], interval="day", span="year", bounds="regular"
            )
            closest = nearest_value(historicals, at_day)
            if closest:
                return Decimal(closest["high_price"])
            raise PriceFetchError(
                f"No historical data found for ticker {ticker} on date {at_day.isoformat()}"
            )
        else:
            local = self._local_latest_price_cache.get(ticker)
            if local:
                return local
            quotes = self._provider.get_quotes([ticker])
            if not quotes[0]:
                return None
            rval = Decimal(quotes[0]["ask_price"])
            self._local_latest_price_cache[ticker] = rval
            return rval

    def _buy_instrument(
        self, symbol: str, qty: float, value: Optional[Money] = None
    ) -> dict:
        """Custom function to enable evolution with the robinhood API"""
        from robin_stocks.robinhood.stocks import (
            orders_url,
            # request_post,
            SESSION,
            update_session,
        )
        from robin_stocks.robinhood.orders import (
            load_account_profile,
            get_latest_price,
            round_price,
        )
        from uuid import uuid4

        def request_post(
            url, payload=None, timeout=16, json=False, jsonify_data=True, retry: int = 1
        ):
            """For a given url and payload, makes a post request and returns the response. Allows for responses other than 200.

            :param url: The url to send a post request to.
            :type url: str
            :param payload: Dictionary of parameters to pass to the url as url/?key1=value1&key2=value2.
            :type payload: Optional[dict]
            :param timeout: The time for the post to wait for a response. Should be slightly greater than multiples of 3.
            :type timeout: Optional[int]
            :param json: This will set the 'content-type' parameter of the session header to 'application/json'
            :type json: bool
            :param jsonify_data: If this is true, will return requests.post().json(), otherwise will return response from requests.post().
            :type jsonify_data: bool
            :returns: Returns the data from the post request.

            """
            data = None
            res = None
            if json:
                update_session("Content-Type", "application/json")
                res = SESSION.post(url, json=payload, timeout=timeout)
                update_session(
                    "Content-Type", "application/x-www-form-urlencoded; charset=utf-8"
                )
            else:
                res = SESSION.post(url, data=payload, timeout=timeout)

            if res.status_code == 429:
                sleep(5 * retry)
                return request_post(
                    url, payload, timeout, json, jsonify_data, retry + 1
                )
            if res.status_code not in [
                200,
                201,
                202,
                204,
                301,
                302,
                303,
                304,
                307,
                400,
                401,
                402,
                403,
            ]:
                res.raise_for_status()

            data = res.json()
            if jsonify_data:
                return data
            else:
                return res

        price = round_price(
            next(iter(get_latest_price(symbol, "ask_price", False)), 0.00)
        )
        if value:
            qty = round(float(value.decimal) / price, 2)

        account = self._get_cached_value(
            CacheKey.MISC,
            value="account_id",
            callable=lambda: load_account_profile(account_number=None, info="url"),
        )
        sym_to_i = self._get_cached_value(
            CacheKey.MISC,
            value="symbol_to_instrument",
            callable=lambda: {
                row["symbol"]: row["url"] for row in self._local_instrument_cache
            },
        )
        payload = {
            "account": account,
            "instrument": sym_to_i[symbol],
            "order_form_version": "2",
            "preset_percent_limit": "0.05",
            "symbol": symbol,
            "price": price,
            "quantity": qty,
            "ref_id": str(uuid4()),
            "type": "limit",
            "time_in_force": "gfd",
            "trigger": "immediate",
            "side": "buy",
            "extended_hours": False,
        }

        url = orders_url()
        data = request_post(url, payload, json=True, jsonify_data=True)

        return data or {}

    def buy_instrument(self, ticker: str, qty: Decimal, value: Optional[Money] = None):
        float_qty = float(qty)
        output = self._buy_instrument(ticker, float_qty, value)
        msg = output.get("detail")
        if msg and "throttled" in msg:
            m = re.search("available in ([0-9]+) seconds", msg)
            if m:
                found = m.group(1)
                t = int(found)
            else:
                t = 30
            Logger.info(f"RH error: was throttled! Sleeping {t}")
            sleep(t)
            return self.buy_instrument(ticker=ticker, qty=qty)
        elif msg and "Too many requests for fractional orders" in msg:
            Logger.info(
                f"RH error: was throttled on fractional orders! Sleeping {FRACTIONAL_SLEEP}"
            )
            sleep(FRACTIONAL_SLEEP)
            return self.buy_instrument(ticker=ticker, qty=qty)
        if not output.get("id"):
            if msg:
                Logger.error(msg)
                raise ValueError(msg)
            Logger.error(output)
            raise ValueError(output)
        return True

    def get_unsettled_instruments(self) -> set[str]:
        from robin_stocks.robinhood.orders import orders_url, request_get

        """We need to efficiently bypass
        paginating all orders if possible
        so just check the account info for if there
        is any cash held for orders first"""
        accounts_data = self._get_cached_value(
            CacheKey.ACCOUNT, callable=self._provider.load_account_profile
        )
        if not accounts_data:
            raise ConfigurationError("Could not load account profile, check login")
        if float(accounts_data.get("cash_held_for_orders", 0)) == 0:
            return set()

        url = orders_url()
        from datetime import datetime, timedelta

        window = datetime.now() - timedelta(days=7)
        data = request_get(url, "results", payload={"updated_at": window.isoformat()})
        orders = [item for item in data if item["cancel"] is not None]
        if len(orders) == len(data):
            # bit the bullet
            data = request_get(
                url, "paginate", payload={"updated_at": window.isoformat()}
            )
        instrument_to_symbol_map = {
            row["url"]: row["symbol"] for row in self._local_instrument_cache
        }
        for item in orders:
            item["symbol"] = instrument_to_symbol_map[item["instrument"]]
        return set(item["symbol"] for item in orders)

    def _refresh_local_instruments(self):
        from robin_stocks.robinhood.stocks import instruments_url, request_get

        instrument_url = instruments_url()
        instrument_info = request_get(instrument_url, dataType="pagination")
        self._local_instrument_cache = instrument_info
        self._save_local_instrument_cache()
        return self._local_instrument_cache

    def _process_cache_to_dict(self):
        return InstrumentDict(
            lambda: {
                row["url"]: row["symbol"] for row in self._refresh_local_instruments()
            },
            {row["url"]: row["symbol"] for row in self._local_instrument_cache},
        )

    def _get_local_instrument_symbol(
        self, instrument: str, refreshed: bool = False
    ) -> str:
        instrument_to_symbol_map = self._get_cached_value(
            CacheKey.MISC,
            value="instrument_to_symbol_map",
            callable=self._process_cache_to_dict,
        )
        try:
            out = instrument_to_symbol_map[instrument]
            return out
        except KeyError as e:
            if not refreshed:
                instrument_to_symbol_map = self._get_cached_value(
                    CacheKey.MISC,
                    value="instrument_to_symbol_map",
                    callable=self._process_cache_to_dict,
                    # force refresh
                    max_age_seconds=1,
                )
                return self._get_local_instrument_symbol(instrument, True)
            raise e

    def _get_stock_info(self, ticker: str) -> dict:
        matches = self._provider.find_instrument_data(ticker)
        for match in matches:
            if match["symbol"] == ticker:
                return {
                    "name": match["simple_name"],
                    "exchange": match["exchange"],
                    "market": match["market"],
                    "country": match["country"],
                    "tradable": bool(match["tradable"]),
                }
        return {}

    def get_holdings(self):
        accounts_data = self._get_cached_value(
            CacheKey.ACCOUNT, callable=self._provider.load_account_profile
        )
        my_stocks = self._get_cached_value(
            CacheKey.POSITIONS, callable=self._provider.get_open_stock_positions
        )
        unsettled = self._get_cached_value(
            CacheKey.UNSETTLED, callable=self.get_unsettled_instruments
        )

        pre = {}
        symbols = []
        instrument_to_symbol_map = self._get_cached_value(
            CacheKey.MISC,
            value="instrument_to_symbol_map",
            callable=self._process_cache_to_dict,
        )
        for row in my_stocks:
            local = {}
            local["units"] = row["quantity"]
            # instrument_data = self._provider.get_instrument_by_url(row["instrument"])
            ticker = instrument_to_symbol_map[row["instrument"]]
            local["ticker"] = ticker
            symbols.append(ticker)
            local["value"] = 0
            local["weight"] = 0
            pre[ticker] = local
        # grab this _after_, in case we had to refresh instruments
        inactive_stocks = {
            row["symbol"]
            for row in self._local_instrument_cache
            if row["state"] == "inactive"
        }
        symbols = [s for s in symbols if s not in inactive_stocks]
        prices = self.get_instrument_prices(symbols)
        self._local_latest_price_cache = {**prices, **self._local_latest_price_cache}
        total_value = Decimal(0.0)
        for s in symbols:
            if not self._local_latest_price_cache[s]:
                continue
            total_value += self._local_latest_price_cache[s] * Decimal(pre[s]["units"])
        final = []
        pl = self.get_per_ticker_profit_or_loss()
        for s in symbols:
            local = pre[s]
            value = Decimal(self._local_latest_price_cache[s] or 0) * Decimal(
                pre[s]["units"]
            )
            local["value"] = Money(value=value)
            local["weight"] = value / total_value
            local["unsettled"] = s in unsettled
            if s in pl:
                local["appreciation"] = pl[s].appreciation
                local["dividends"] = pl[s].dividends
            final.append(local)
        out = [RealPortfolioElement(**row) for row in final]

        cash = Decimal(accounts_data["portfolio_cash"]) - Decimal(
            accounts_data["cash_held_for_orders"]
        )
        return RealPortfolio(holdings=out, cash=Money(value=cash), provider=self)

    def get_instrument_prices(self, tickers: List[str], at_day: Optional[date] = None):
        return self._price_cache.get_prices(tickers=tickers, date=at_day)

    def _get_instrument_prices(
        self, tickers: List[str], at_day: Optional[date] = None
    ) -> Dict[str, Optional[Decimal]]:
        ticker_list = tickers
        batches = []
        for batch in divide_into_batches(ticker_list, BATCH_SIZE):
            if at_day:
                historicals = self._provider.get_stock_historicals(
                    batch, interval="day", span="year", bounds="regular"
                )
                batches.append(
                    {s: nearest_multi_value(s, historicals, at_day) for s in batch}
                )
            else:
                results = self._provider.get_quotes(batch)
                batches.append({s: nearest_multi_value(s, results) for s in batch})
        prices: Dict[str, Optional[Decimal]] = {}
        for fbatch in batches:
            prices = {**prices, **fbatch}
        return prices

    def get_per_ticker_profit_or_loss(self) -> Dict[str, ProfitModel]:
        my_stocks = self._get_cached_value(
            CacheKey.POSITIONS, callable=self._provider.get_open_stock_positions
        )
        instrument_to_symbol_map = self._get_cached_value(
            CacheKey.MISC,
            callable=self._process_cache_to_dict,
        )
        divs = self._get_dividends()
        output = {}
        for x in my_stocks:
            historical_value = Decimal(x["average_buy_price"]) * Decimal(x["quantity"])
            ticker = instrument_to_symbol_map[x["instrument"]]
            try:
                current_price = self.get_instrument_price(ticker) or Decimal(0.0)
            except PriceFetchError:
                current_price = Decimal(0.0)
            current_value = current_price * Decimal(x["quantity"])
            pl = Money(value=current_value - historical_value)
            output[ticker] = ProfitModel(
                appreciation=pl, dividends=divs.get(ticker, Money(value=Decimal(0)))
            )
        return output

    def _get_dividends(self) -> DefaultDict[str, Money]:
        from collections import defaultdict

        value = self._get_cached_value(
            CacheKey.DIVIDENDS, callable=self._provider.get_dividends
        )
        output: DefaultDict[str, Money] = defaultdict(lambda: Money(value=0))
        instrument_to_symbol_map = self._get_cached_value(
            CacheKey.MISC,
            callable=self._process_cache_to_dict,
        )
        [
            item.update({"symbol": instrument_to_symbol_map[item["instrument"]]})
            for item in value
        ]
        for item in value:
            if item["state"] == "paid":
                output[item["symbol"]] += Money(value=float(item["amount"]))
        return output



from os.path import dirname, join
from pydantic import BaseModel, Field
from typing import List, Set
from pathlib import Path


class StocklistInventory(BaseModel):
    keys: Set[str] = Field(exclude=True)
    base: Path = Field(exclude=True)
    loaded: dict[str, List[str]] = Field(default_factory=dict)

    @classmethod
    def from_path(cls, path):
        path = Path(path)
        if path.is_file():
            path = path.parent
        keys = []
        for f in path.iterdir():
            try:
                if f.suffix == ".csv":
                    keys.append(f.stem)
            except FileNotFoundError:
                pass
        return StocklistInventory(keys=keys, base=path)

    def __getitem__(self, item: str) -> List[str]:
        if item in self.keys:
            values = self.loaded.get(item, None)
            if values:
                return values
            values = self.get_values(item)
            self.loaded[item] = values
            return values
        raise KeyError(item)

    def get_values(self, item: str) -> List[str]:
        out = []
        with open(join(dirname(__file__), f"{item}.csv")) as f:
            contents = f.read()
            for row in contents.split("\n"):
                ticker = row.strip()
                out.append(ticker)
        return out

    def add_list(self, key: str, ticker_list: List[str]):
        """Add a new list or merge into existing list."""
        self.keys.add(key)
        current = self.loaded.get(key, [])
        self.loaded[key] = current + ticker_list



from py_portfolio_index.models import RealPortfolio, RealPortfolioElement
from .base_portfolio import BaseProvider
from decimal import Decimal
from typing import Optional
from datetime import date, datetime, timezone, timedelta
from functools import lru_cache
from py_portfolio_index.models import Money
from os import environ


class AlpacaProviderLegacy(BaseProvider):
    def __init__(
        self,
        key_id: str | None = None,
        secret_key: str | None = None,
        paper: bool = False,
    ):
        import alpaca_trade_api as tradeapi
        from alpaca_trade_api.common import URL

        if not key_id:
            key_id = environ.get("ALPACA_API_KEY", None)
        if not secret_key:
            secret_key = environ.get("ALPACA_API_SECRET", None)
        if not (key_id and secret_key):
            raise ValueError(
                "Must provide key_id and secret_key or set environment variables ALPACA_API_KEY and ALPACA_API_SECRET "
            )
        TARGET_URL = (
            "https://paper-api.alpaca.markets"
            if paper
            else "https://api.alpaca.markets"
        )
        self.api = tradeapi.REST(
            key_id=key_id, secret_key=secret_key, base_url=URL(TARGET_URL)
        )
        BaseProvider.__init__(self)

    @lru_cache(maxsize=None)
    def _get_instrument_price(
        self, ticker: str, at_day: Optional[date] = None
    ) -> Optional[Decimal]:
        from alpaca_trade_api.rest import TimeFrame, TimeFrameUnit

        if at_day:
            start = datetime(at_day.year, at_day.month, at_day.day, tzinfo=timezone.utc)
            end = start + timedelta(days=7)
            # end = datetime(at_day.year, at_day.month, at_day.day+7, hour=23, tzinfo=timezone.utc)
            raw = self.api.get_bars(
                [ticker],
                timeframe=TimeFrame(amount=1, unit=TimeFrameUnit.Day),
                start=start.isoformat(),
                end=end.isoformat(),
            )
            # take the first day after target day
            return Decimal(raw[0].h)
        else:
            raw = self.api.get_latest_quote(ticker)
            # if we don't have a value for the current day
            # expand out
            # this is requuired on weekends
            if not raw.ap:
                default = datetime.now(tz=timezone.utc) - timedelta(hours=1)
                start = default - timedelta(days=7)
                end = default
                raw = self.api.get_bars(
                    [ticker],
                    timeframe=TimeFrame(amount=1, unit=TimeFrameUnit.Day),
                    start=start.isoformat(),
                    end=end.isoformat(),
                )
                # take the first day after target day
                return Decimal(raw[0].h)
            return Decimal(raw.ap)

    def buy_instrument(self, ticker: str, qty: Decimal, value: Optional[Money] = None):
        qty_float = float(qty)
        self.api.submit_order(
            symbol=ticker,
            qty=qty_float,
            side="buy",
            type="market",
            time_in_force="day",
        )
        return True

    def get_unsettled_instruments(self):
        open_orders = self.api.list_orders(
            status="open", limit=100, nested=True  # show nested multi-leg orders
        )
        return set([o.symbol for o in open_orders])

    def get_holdings(self):
        from decimal import Decimal

        my_stocks = self.api.list_positions()
        # df = pd.DataFrame(my_stocks)
        if not my_stocks:
            return RealPortfolio(holdings=[])
        total_value = sum([Decimal(item.market_value) for item in my_stocks])
        out = [
            RealPortfolioElement(
                ticker=row.symbol,
                units=row.qty,
                value=Decimal(row.market_value),
                weight=Decimal(row.market_value) / total_value,
            )
            for row in my_stocks
        ]
        return RealPortfolio(holdings=out)
